from abc import ABC, abstractmethod
from collections.abc import Iterator
from typing import final

from .node import Node, iter_node
from .view import View


class Component(ABC, View):
    @abstractmethod
    def render(self) -> Node: ...

    @final
    def __iter__(self) -> Iterator[str]:
        yield from iter_node(self.render())



from functools import lru_cache
from re import match as re_fullmatch
from re import sub as re_sub

from ._private.element import CommentElement, Element, HtmlElement, VoidElement

__all__ = [
    "_",
    "Html",
    "Area",
    "Base",
    "Br",
    "Col",
    "Embed",
    "Hr",
    "Img",
    "Input",
    "Link",
    "Meta",
    "Param",
    "Source",
    "Track",
    "Wbr",
    "A",
    "Abbr",
    "Abc",
    "Address",
    "Article",
    "Aside",
    "Audio",
    "B",
    "Bdi",
    "Bdo",
    "Blockquote",
    "Body",
    "Button",
    "Canvas",
    "Caption",
    "Cite",
    "Code",
    "Colgroup",
    "Data",
    "Datalist",
    "Dd",
    "Del",
    "Details",
    "Dfn",
    "Dialog",
    "Div",
    "Dl",
    "Dt",
    "Em",
    "Fieldset",
    "Figcaption",
    "Figure",
    "Footer",
    "Form",
    "H1",
    "H2",
    "H3",
    "H4",
    "H5",
    "H6",
    "Head",
    "Header",
    "Hgroup",
    "I",
    "Iframe",
    "Ins",
    "Kbd",
    "Label",
    "Legend",
    "Li",
    "Main",
    "Map",
    "Mark",
    "Menu",
    "Meter",
    "Nav",
    "Noscript",
    "Object",
    "Ol",
    "Optgroup",
    "Option",
    "Output",
    "P",
    "Picture",
    "Portal",
    "Pre",
    "Progress",
    "Q",
    "Rp",
    "Rt",
    "Ruby",
    "S",
    "Samp",
    "Script",
    "Search",
    "Section",
    "Select",
    "Slot",
    "Small",
    "Span",
    "Strong",
    "Style",
    "Sub",
    "Summary",
    "Sup",
    "Table",
    "Tbody",
    "Td",
    "Template",
    "Textarea",
    "Tfoot",
    "Th",
    "Thead",
    "Time",
    "Title",
    "Tr",
    "U",
    "Ul",
    "Var",
]


@lru_cache(maxsize=300)
def _get_element(name: str) -> Element:
    if not re_fullmatch(r"^(?:[A-Z][a-z]*)+$", name):
        raise AttributeError(
            f"{name} is not a valid element name. markupy tags must be in CapitalizedCase"
        )

    #  Uppercase chars are word boundaries for tag names
    words = filter(None, re_sub(r"([A-Z])", r" \1", name).split())
    html_name = "-".join(words).lower()
    return Element(html_name)


def __getattr__(name: str) -> Element:
    return _get_element(name)


_ = CommentElement("_")

# https://developer.mozilla.org/en-US/docs/Glossary/Doctype
Html = HtmlElement("html")

# https://developer.mozilla.org/en-US/docs/Glossary/Void_element
Area = VoidElement("area")
Base = VoidElement("base")
Br = VoidElement("br")
Col = VoidElement("col")
Embed = VoidElement("embed")
Hr = VoidElement("hr")
Img = VoidElement("img")
Input = VoidElement("input")
Link = VoidElement("link")
Meta = VoidElement("meta")
Param = VoidElement("param")
Source = VoidElement("source")
Track = VoidElement("track")
Wbr = VoidElement("wbr")

# Non-deprecated HTML elements, extracted from
# https://developer.mozilla.org/en-US/docs/Web/HTML/Element
# Located via the inspector with:
# Array.from($0.querySelectorAll('li')).filter(x=>!x.querySelector('.icon-deprecated')).map(x => x.querySelector('code').textContent) # noqa: E501
A = Element("a")
Abbr = Element("abbr")
Abc = Element("abc")
Address = Element("address")
Article = Element("article")
Aside = Element("aside")
Audio = Element("audio")
B = Element("b")
Bdi = Element("bdi")
Bdo = Element("bdo")
Blockquote = Element("blockquote")
Body = Element("body")
Button = Element("button")
Canvas = Element("canvas")
Caption = Element("caption")
Cite = Element("cite")
Code = Element("code")
Colgroup = Element("colgroup")
Data = Element("data")
Datalist = Element("datalist")
Dd = Element("dd")
Del = Element("del")
Details = Element("details")
Dfn = Element("dfn")
Dialog = Element("dialog")
Div = Element("div")
Dl = Element("dl")
Dt = Element("dt")
Em = Element("em")
Fieldset = Element("fieldset")
Figcaption = Element("figcaption")
Figure = Element("figure")
Footer = Element("footer")
Form = Element("form")
H1 = Element("h1")
H2 = Element("h2")
H3 = Element("h3")
H4 = Element("h4")
H5 = Element("h5")
H6 = Element("h6")
Head = Element("head")
Header = Element("header")
Hgroup = Element("hgroup")
I = Element("i")  # noqa: E741
Iframe = Element("iframe")
Ins = Element("ins")
Kbd = Element("kbd")
Label = Element("label")
Legend = Element("legend")
Li = Element("li")
Main = Element("main")
Map = Element("map")
Mark = Element("mark")
Menu = Element("menu")
Meter = Element("meter")
Nav = Element("nav")
Noscript = Element("noscript")
Object = Element("object")
Ol = Element("ol")
Optgroup = Element("optgroup")
Option = Element("option")
Output = Element("output")
P = Element("p")
Picture = Element("picture")
Portal = Element("portal")
Pre = Element("pre")
Progress = Element("progress")
Q = Element("q")
Rp = Element("rp")
Rt = Element("rt")
Ruby = Element("ruby")
S = Element("s")
Samp = Element("samp")
Script = Element("script")
Search = Element("search")
Section = Element("section")
Select = Element("select")
Slot = Element("slot")
Small = Element("small")
Span = Element("span")
Strong = Element("strong")
Style = Element("style")
Sub = Element("sub")
Summary = Element("summary")
Sup = Element("sup")
Table = Element("table")
Tbody = Element("tbody")
Td = Element("td")
Template = Element("template")
Textarea = Element("textarea")
Tfoot = Element("tfoot")
Th = Element("th")
Thead = Element("thead")
Time = Element("time")
Title = Element("title")
Tr = Element("tr")
U = Element("u")
Ul = Element("ul")
Var = Element("var")



from markupy import Component, Node, tag


class ComponentElement(Component):
    def render(self) -> Node:
        return tag.Div


class ComponentElementMultiple(Component):
    def render(self) -> Node:
        return tag.Div, tag.Img


class ComponentInComponent(Component):
    def render(self) -> Node:
        return tag.Input, tag.Div("#parent")[ComponentElement()]


def test_component_element() -> None:
    assert str(ComponentElement()) == """<div></div>"""


def test_component_element_multiple() -> None:
    assert str(ComponentElementMultiple()) == """<div></div><img>"""


def test_component_in_component() -> None:
    assert (
        str(ComponentInComponent()) == """<input><div id="parent"><div></div></div>"""
    )



from markupsafe import Markup

from markupy._private.element import (
    CommentElement,
    Element,
    HtmlElement,
    VoidElement,
)
from markupy.tag import Div, Hr, Html, Input, MyElement, Unknown, _


def test_instance_cache() -> None:
    """
    markupy creates element object dynamically. make sure they are reused.
    """
    from markupy import tag

    assert tag.Div is tag.Div
    assert tag.Div is tag.Div()
    assert tag.Div is tag.Div(attr=False)
    assert tag.Div is tag.Div[None]
    assert tag.Div is tag.Div[True]
    assert tag.Div is tag.Div[False]
    assert tag.Div is tag.Div[[]]
    assert tag.Div is tag.Div[[None]]
    assert tag.Div is tag.Div[""]
    assert tag.Div is not tag.Div[0]


def test_element_repr() -> None:
    assert repr(Div("#a")) == """<Element '<div id="a">'>"""


def test_void_element_repr() -> None:
    assert repr(Hr("#a")) == """<VoidElement '<hr id="a">'>"""


def test_markup_str() -> None:
    result = str(Div(id="a"))
    assert isinstance(result, str)
    assert isinstance(result, Markup)
    assert result == '<div id="a"></div>'


def test_element_type() -> None:
    assert type(Unknown) is Element
    assert type(MyElement) is Element
    assert type(Div) is Element
    assert type(Input) is VoidElement
    assert type(Html) is HtmlElement
    assert type(_) is CommentElement


def test_comment() -> None:
    assert str(_["Hello"]) == "<!--Hello-->"
    assert str(_[Div["Hello"]]) == "<!--<div>Hello</div>-->"



from collections.abc import Iterator
from typing import Any, overload

from typing_extensions import Self, override

from .attribute import AttributeDict, AttributeValue
from .component import Component
from .node import Node, iter_node, validate_node
from .view import View


@Component.register
class Element(View):
    def __init__(self, name: str, *, shared: bool = True) -> None:
        self._name = name
        self._attributes: str | None = None
        self._children: Node = None
        self._shared: bool = shared

    @property
    def name(self) -> str:
        return self._name

    def _tag_opening(self) -> str:
        if attributes := self._attributes:
            return f"<{self._name} {attributes}>"
        return f"<{self._name}>"

    def _tag_closing(self) -> str:
        return f"</{self._name}>"

    def __iter__(self) -> Iterator[str]:
        yield self._tag_opening()
        yield from iter_node(self._children)
        yield self._tag_closing()

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} '{self._tag_opening()}'>"

    def _new_instance(self: Self) -> Self:
        # When imported, elements are loaded from a shared instance
        # Make sure we re-instantiate them on setting attributes/children
        # to avoid sharing attributes/children between multiple instances
        if self._shared:
            return self.__class__(self._name, shared=False)
        return self

    # Use call syntax () to define attributes
    @overload
    def __call__(
        self,
        selector: str,
        attributes: dict[str, AttributeValue],
        **kwargs: AttributeValue,
    ) -> Self: ...
    @overload
    def __call__(self, selector: str, **kwargs: AttributeValue) -> Self: ...
    @overload
    def __call__(
        self, attributes: dict[str, AttributeValue], **kwargs: AttributeValue
    ) -> Self: ...
    @overload
    def __call__(self, **kwargs: AttributeValue) -> Self: ...
    def __call__(self, *args: Any, **kwargs: Any) -> Self:
        selector: str | None = None
        attributes_dict: dict[str, AttributeValue] | None = None
        attributes_kwargs: dict[str, AttributeValue] = kwargs
        if len(args) == 1:
            arg = args[0]
            if isinstance(arg, str):
                # element(".foo")
                selector = arg
            elif isinstance(arg, dict):
                # element({"foo": "bar"})
                attributes_dict = arg
            else:
                raise TypeError(
                    f"Invalid argument type `{arg!r}` for element {self}, expected `str` or `dict`"
                )
        elif len(args) == 2:
            # element(".foo", {"bar": "baz"})
            if not isinstance(args[0], str):
                raise TypeError(
                    f"Invalid first argument type `{args[0]!r}` for element {self}, expected `str`"
                )
            if not isinstance(args[1], dict):
                raise TypeError(
                    f"Invalid second argument type `{args[1]!r}` for element {self}, expected `dict`"
                )
            selector, attributes_dict = args
        elif len(args) > 2:
            raise ValueError(f"Invalid number of arguments provided for element {self}")

        if not selector and not attributes_dict and not attributes_kwargs:
            return self

        attrs = AttributeDict()
        try:
            attrs.add_selector(selector)
        except (TypeError, ValueError):
            raise ValueError(f"Invalid selector string `{selector}` for element {self}")
        try:
            attrs.add_dict(attributes_dict)
        except (TypeError, ValueError):
            raise ValueError(
                f"Invalid dict attributes `{attributes_dict}` for element {self}"
            )
        try:
            attrs.add_dict(attributes_kwargs, rewrite_keys=True)
        except (TypeError, ValueError):
            raise ValueError(
                f"Invalid keyword attributes `{attributes_kwargs}` for element {self}"
            )

        if attributes := str(attrs):
            el = self._new_instance()
            el._attributes = attributes
            return el

        return self

    # Use subscriptable [] syntax to assign children
    def __getitem__(self, children: "Node") -> Self:
        if validate_node(children):
            el = self._new_instance()
            el._children = children
            return el

        return self


class HtmlElement(Element):
    @override
    def __iter__(self) -> Iterator[str]:
        yield "<!doctype html>"
        yield from super().__iter__()


class VoidElement(Element):
    @override
    def __iter__(self) -> Iterator[str]:
        yield self._tag_opening()

    @override
    def __getitem__(self, children: Any) -> Self:
        raise ValueError(f"Void element {self} cannot contain children")


class CommentElement(Element):
    @override
    def _tag_opening(self) -> str:
        return "<!--"

    @override
    def _tag_closing(self) -> str:
        return "-->"

    @override
    def __call__(self, *args: Any, **kwargs: Any) -> Self:
        raise ValueError(f"Comment element {self} cannot have attributes")



from ._private.component import Component
from ._private.html import to_markupy as html2markupy
from ._private.node import Node, iter_node, render_node

__all__ = [
    "Node",
    "Component",
    "html2markupy",
    "iter_node",
    "render_node",
]
__version__ = "1.0.0"



from collections.abc import Iterable
from typing import final

from markupsafe import Markup


class View(Iterable[str]):
    @final
    def __str__(self) -> str:
        return Markup("".join(self))

    # Allow starlette Response.render to directly render this element without
    # explicitly casting to str:
    # https://github.com/encode/starlette/blob/5ed55c441126687106109a3f5e051176f88cd3e6/starlette/responses.py#L44-L49
    def encode(self, encoding: str = "utf-8", errors: str = "strict") -> bytes:
        return str(self).encode(encoding, errors)

    # Avoid having Django "call" a markupy element that is injected into a
    # template. Setting do_not_call_in_templates will prevent Django from doing
    # an extra call:
    # https://docs.djangoproject.com/en/5.0/ref/templates/api/#variables-and-lookups
    do_not_call_in_templates = True
